# Copyright (c) 2019-2022, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""EEPROM management for PyFdti"""

#pylint: disable-msg=too-many-arguments
#pylint: disable-msg=too-many-branches
#pylint: disable-msg=too-many-instance-attributes
#pylint: disable-msg=too-many-locals
#pylint: disable-msg=too-many-public-methods
#pylint: disable-msg=wrong-import-position
#pylint: disable-msg=import-error

import sys
from binascii import hexlify, unhexlify
from collections import OrderedDict, namedtuple
from configparser import ConfigParser
from enum import IntEnum
if sys.version_info[:2] > (3, 5):
    from enum import IntFlag
from logging import getLogger
from random import randint
from re import match
from struct import calcsize as scalc, pack as spack, unpack as sunpack
from typing import BinaryIO, List, Optional, Set, TextIO, Union, Tuple
if sys.version_info[:2] == (3, 5):
    from aenum import IntFlag
from usb.core import Device as UsbDevice
from .ftdi import Ftdi, FtdiError
from .misc import classproperty, to_bool, to_int


class FtdiEepromError(FtdiError):
    """FTDI EEPROM error."""


class Hex2Int(int):
    """Hexa representation of a byte."""
    def __str__(self):
        return '0x%02x' % int(self)


class Hex4Int(int):
    """Hexa representation of a half-word."""
    def __str__(self):
        return '0x%04x' % int(self)


class FtdiEeprom:
    """FTDI EEPROM management
    """

    _PROPS = namedtuple('PROPS', 'size user dynoff chipoff')
    """Properties for each FTDI device release.

       * size is the size in bytes of the EEPROM storage area
       * user is the size in bytes of the user storage area, if any/supported
       * dynoff is the offset in EEPROM of the first bytes to store strings
       * chipoff is the offset in EEPROM of the EEPROM chip type
    """

    _PROPERTIES = {
        0x0200: _PROPS(0, None, 0, None),        # FT232AM
        0x0400: _PROPS(256, 0x14, 0x94, None),   # FT232BM
        0x0500: _PROPS(256, 0x16, 0x96, 0x14),   # FT2232D
        0x0600: _PROPS(128, None, 0x18, None),   # FT232R
        0x0700: _PROPS(256, 0x1A, 0x9A, 0x18),   # FT2232H
        0x0800: _PROPS(256, 0x1A, 0x9A, 0x18),   # FT4232H
        0x0900: _PROPS(256, 0x1A, 0xA0, 0x1e),   # FT232H
        0x1000: _PROPS(1024, 0x1A, 0xA0, None),  # FT230X/FT231X/FT234X
    }
    """EEPROM properties."""


    CBUS = IntEnum('CBus',
                   'TXDEN PWREN TXLED RXLED TXRXLED SLEEP CLK48 CLK24 CLK12 '
                   'CLK6 GPIO BB_WR BB_RD', start=0)
    """Alternate features for legacy FT232R devices."""

    CBUSH = IntEnum('CBusH',
                    'TRISTATE TXLED RXLED TXRXLED PWREN SLEEP DRIVE0 DRIVE1 '
                    'GPIO TXDEN CLK30 CLK15 CLK7_5', start=0)
    """Alternate features for FT232H/FT2232H/FT4232H devices."""

    CBUSX = IntEnum('CBusX',
                    'TRISTATE TXLED RXLED TXRXLED PWREN SLEEP DRIVE0 DRIVE1 '
                    'GPIO TXDEN CLK24 CLK12 CLK6 BAT_DETECT BAT_NDETECT '
                    'I2C_TXE I2C_RXF VBUS_SENSE BB_WR BB_RD TIMESTAMP AWAKE',
                    start=0)
    """Alternate features for FT230X devices."""

    UART_BITS = IntFlag('UartBits', 'TXD RXD RTS CTS DTR DSR DCD RI')
    """Inversion flags for FT232R and FT-X devices."""

    CHANNEL = IntFlag('Channel', 'FIFO OPTO CPU FT128 RS485')
    """Alternate port mode."""

    DRIVE = IntFlag('Drive',
                    'LOW HIGH SLOW_SLEW SCHMITT _10 _20 _40 PWRSAVE_DIS')
    """Driver options for I/O pins."""

    CFG1 = IntFlag('Cfg1', 'CLK_IDLE_STATE DATA_LSB FLOW_CONTROL _08 '
                           'HIGH_CURRENTDRIVE _20 _40 SUSPEND_DBUS7')
    """Configuration bits stored @ 0x01."""

    VAR_STRINGS = ('manufacturer', 'product', 'serial')
    """EEPROM strings with variable length."""

    def __init__(self):
        self.log = getLogger('pyftdi.eeprom')
        self._ftdi = Ftdi()
        self._eeprom = bytearray()
        self._size = 0
        self._dev_ver = 0
        self._valid = False
        self._config = OrderedDict()
        self._dirty = set()
        self._modified = False
        self._chip: Optional[int] = None
        self._mirror = False

    def __getattr__(self, name):
        if name in self._config:
            return self._config[name]
        raise AttributeError('No such attribute: %s' % name)

    @classproperty
    def eeprom_sizes(cls) -> List[int]:
        """Return a list of supported EEPROM sizes.

           :return: the supported EEPROM sizes
        """
        return sorted({p.size for p in cls._PROPERTIES.values() if p.size})

    def open(self, device: Union[str, UsbDevice],
             ignore: bool = False, size: Optional[int] = None,
             model: Optional[str] = None) -> None:
        """Open a new connection to the FTDI USB device.

           :param device: the device URL or a USB device instance.
           :param ignore: whether to ignore existing content
           :param size: a custom EEPROM size
           :param model: the EEPROM model used to specify a custom size
        """
        if self._ftdi.is_connected:
            raise FtdiError('Already open')
        if isinstance(device, str):
            self._ftdi.open_from_url(device)
        else:
            self._ftdi.open_from_device(device)
        if model and not size:
            # 93xxx46/56/66
            mmo = match(r'(?i)^93[a-z]*([456])6.*$', model)
            if not mmo:
                raise ValueError(f'Unknown EEPROM device: {model}')
            mmul = int(mmo.group(1))
            size = 128 << (mmul - 4)
        if size:
            if size not in self.eeprom_sizes:
                raise ValueError(f'Unsupported EEPROM size: {size}')
            self._size = min(size, 256)
        if not ignore:
            self._eeprom = self._read_eeprom()
            if self._valid:
                self._decode_eeprom()

    def close(self) -> None:
        """Close the current connection to the FTDI USB device,
        """
        if self._ftdi.is_connected:
            self._ftdi.close()
            self._eeprom = bytearray()
            self._dev_ver = 0
            self._config.clear()

    def connect(self, ftdi: Ftdi, ignore: bool = False) -> None:
        """Connect a FTDI EEPROM to an existing Ftdi instance.

           :param ftdi: the Ftdi instance to use
           :param ignore: whether to ignore existing content
        """
        self._ftdi = ftdi
        self._eeprom = bytearray()
        self._dev_ver = 0
        self._valid = False
        self._config = OrderedDict()
        self._dirty = set()
        if not ignore:
            self._eeprom = self._read_eeprom()
            if self._valid:
                self._decode_eeprom()
            self._decode_eeprom()

    @property
    def device_version(self) -> int:
        """Report the version of the FTDI device.

           :return: the release
        """
        if not self._dev_ver:
            if not self._ftdi.is_connected:
                raise FtdiError('Not connected')
            self._dev_ver = self._ftdi.device_version
        return self._dev_ver

    @property
    def size(self) -> int:
        """Report the EEPROM size.

           Use the most common (default) EEPROM size of the size is not yet
           known.

           :return: the size in bytes
        """
        if not self._size:
            self._size = self.default_size
        return self._size

    @property
    def default_size(self) -> int:
        """Report the default EEPROM size based on the FTDI type.

           The physical EEPROM size may be greater or lower, depending on the
           actual connected EEPROM device.

           :return: the size in bytes
        """
        if self._chip == 0x46:
            return 0x80  # 93C46
        if self._chip == 0x56:
            return 0x100  # 93C56
        if self._chip == 0x66:
            return 0x100  # 93C66 (512 bytes, only 256 are used)
        try:
            eeprom_size = self._PROPERTIES[self.device_version].size
        except (AttributeError, KeyError) as exc:
            raise FtdiError('No EEPROM') from exc
        return eeprom_size

    @property
    def storage_size(self) -> int:
        """Report the number of EEPROM bytes that can be used for configuration
            storage. The physical EEPROM size may be greater

            :return: the number of bytes in the eeprom that will be used for
                configuration storage
        """
        try:
            eeprom_storage_size = self.size
            if self.is_mirroring_enabled:
                eeprom_storage_size = self.mirror_sector
        except FtdiError as fe:
            raise fe
        return eeprom_storage_size

    @property
    def data(self) -> bytes:
        """Returns the content of the EEPROM.

           :return: the content as bytes.
        """
        self._sync_eeprom()
        return bytes(self._eeprom)

    @property
    def properties(self) ->  Set[str]:
        """Returns the supported properties for the current device.

           :return: the supported properies.
        """
        props = set(self._config.keys())
        props -= set(self.VAR_STRINGS)
        return props

    @property
    def is_empty(self) -> bool:
        """Reports whether the EEPROM has been erased, or no EEPROM is
           connected to the FTDI EEPROM port.

           :return: True if no content is detected
        """
        if len(self._eeprom) != self.size:
            return False
        for byte in self._eeprom:
            if byte != 0xFF:
                return False
        return True

    @property
    def cbus_pins(self) -> List[int]:
        """Return the list of CBUS pins configured as GPIO, if any

           :return: list of CBUS pins
        """
        pins = [pin for pin in range(0, 10)
                if self._config.get('cbus_func_%d' % pin, '') == 'GPIO']
        return pins

    @property
    def cbus_mask(self) -> int:
        """Return the bitmask of CBUS pins configured as GPIO.

           The bitmap contains four bits, ordered in natural order.

           :return: CBUS mask
        """
        if self.device_version == 0x900:  # FT232H
            cbus = [5, 6, 8, 9]
        else:
            cbus = list(range(4))
        mask = 0
        for bix, pin in enumerate(cbus):
            if self._config.get('cbus_func_%d' % pin, '') == 'GPIO':
                mask |= 1 << bix
        return mask

    @property
    def has_mirroring(self) -> bool:
        """Report whether the device supports EEPROM content duplication
           across its two sectors.

           :return: True if the device support mirorring
        """
        return (self._PROPERTIES[self.device_version].user and
                self._ftdi.device_version != 0x1000)

    @property
    def mirror_sector(self) -> int:
        """Report start address of the mirror sector in the EEPROM.
           This is only valid if the FTDI is capable of mirroring EEPROM data.

           :return: the start address
        """
        if self.has_mirroring:
            return self.size // 2
        raise FtdiError('EEPROM does not support mirroring')

    @property
    def is_mirroring_enabled(self) -> bool:
        """Check if EEPROM mirroring is currently enabled for this EEPROM.
            See enable_mirroring for more details on EEPROM mirroring
            functionality
        """
        return self.has_mirroring and self._mirror

    def enable_mirroring(self, enable : bool) -> None:
        """Enable EEPROM write mirroring. When enabled, this divides the EEPROM
           into 2 sectors and mirrors configuration data between them.

           For example on a 256 byte EEPROM, two 128 byte 'sectors' will be
           used to store identical data. Configuration properties/strings will
           be writen to both of these sectors. For some devices (like the
           4232H), this makes the PyFtdi EEPROM functionally similar to
           FT_PROG.

           Note: Data will only be mirrored if the has_mirroring property
           returns true (after establishing a connection to the ftdi)

           :param enable: enable or disable EEPROM mirroring
        """
        self._mirror = enable

    def save_config(self, file: TextIO) -> None:
        """Save the EEPROM content as an INI stream.

           :param file: output stream
        """
        self._sync_eeprom()
        cfg = ConfigParser()
        cfg.add_section('values')
        for name, value in self._config.items():
            val = str(value)
            if isinstance(value, bool):
                val = val.lower()
            cfg.set('values', name, val)
        cfg.add_section('raw')
        length = 16
        for i in range(0, len(self._eeprom), length):
            chunk = self._eeprom[i:i+length]
            hexa = hexlify(chunk).decode()
            cfg.set('raw', '@%02x' % i, hexa)
        cfg.write(file)

    def load_config(self, file: TextIO, section: Optional[str] = None) -> None:
        """Load the EEPROM content from an INI stream.

           The ``section`` argument selects which section(s) to load:

           * ``raw`` only loads the raw data (hexabytes) from a previous dump
           * ``values`` only loads the values section, that is the human
             readable configuration.
           * ``all``, which is the default section selection, load the raw
             section, then overwrite part of it with any configuration value
             from the ``values`` section. This provides a handy way to use an
             existing dump from a valid EEPROM content, while customizing some
             parameters, such as the serial number.

           :param file: input stream
           :paran section: which section to load from the ini file
        """
        self._sync_eeprom()
        cfg = ConfigParser()
        cfg.read_file(file)
        loaded = False
        sections = cfg.sections()
        if section not in ('all', None) and section not in sections:
            raise FtdiEepromError(f'No such configuration section {section}')
        sect = 'raw'
        if sect in sections and section in (None, 'all', sect):
            if not cfg.has_section(sect):
                raise FtdiEepromError("No '%s' section in INI file" % sect)
            options = cfg.options(sect)
            try:
                for opt in options:
                    if not opt.startswith('@'):
                        raise ValueError()
                    address = int(opt[1:], 16)
                    hexval = cfg.get(sect, opt).strip()
                    buf = unhexlify(hexval)
                    self._eeprom[address:address+len(buf)] = buf
            except IndexError as exc:
                raise ValueError("Invalid address in '%s'' section" %
                                 sect) from exc
            except ValueError as exc:
                raise ValueError("Invalid line in '%s'' section" %
                                 sect) from exc
            self._compute_crc(self._eeprom, True)
            if not self._valid:
                raise ValueError('Loaded RAW section is invalid (CRC mismatch')
            loaded = True
        sect = 'values'
        vmap = {
            'manufacturer': 'manufacturer_name',
            'product': 'product_name',
            'serial': 'serial_number'
        }
        if sect in sections and section in (None, 'all', sect):
            if not cfg.has_section(sect):
                raise FtdiEepromError("No '%s' section in INI file" % sect)
            options = cfg.options(sect)
            for opt in options:
                value = cfg.get(sect, opt).strip()
                if opt in vmap:
                    func = getattr(self, 'set_%s' % vmap[opt])
                    func(value)
                else:
                    self.log.debug('Assigning opt %s = %s', opt, value)
                    try:
                        self.set_property(opt, value)
                    except (TypeError, ValueError, NotImplementedError) as exc:
                        self.log.warning("Ignoring setting '%s': %s", opt, exc)
            loaded = True
        if not loaded:
            raise ValueError('Invalid section: %s' % section)
        self._sync_eeprom()

    def set_serial_number(self, serial: str) -> None:
        """Define a new serial number."""
        self._validate_string(serial)
        self._update_var_string('serial', serial)
        self.set_property('has_serial', True)

    def set_manufacturer_name(self, manufacturer: str) -> None:
        """Define a new manufacturer string."""
        self._validate_string(manufacturer)
        self._update_var_string('manufacturer', manufacturer)

    def set_product_name(self, product: str) -> None:
        """Define a new product name."""
        self._validate_string(product)
        self._update_var_string('product', product)

    def set_property(self, name: str, value: Union[str, int, bool],
                     out: Optional[TextIO] = None) -> None:
        """Change the value of a stored property.

           :see: :py:meth:`properties` for a list of valid property names.
                 Note that for now, only a small subset of properties can be
                 changed.
           :param name: the property to change
           :param value: the new value (supported values depend on property)
           :param out: optional output stream to report hints
        """
        mobj = match(r'cbus_func_(\d)', name)
        if mobj:
            if not isinstance(value, str):
                raise ValueError("'%s' should be specified as a string" % name)
            self._set_cbus_func(int(mobj.group(1)), value, out)
            self._dirty.add(name)
            return
        mobj = match(r'([abcd])bus_(drive|slow_slew|schmitt)', name)
        if mobj:
            self._set_bus_control(mobj.group(1), mobj.group(2), value, out)
            self._dirty.add(name)
            return
        mobj = match(r'group_(\d)_(drive|schmitt|slow_slew)', name)
        if mobj:
            self._set_group(int(mobj.group(1)), mobj.group(2), value, out)
            self._dirty.add(name)
            return
        confs = {
            'remote_wakeup': (0, 5),
            'self_powered': (0, 6),
            'in_isochronous': (2, 0),
            'out_isochronous': (2, 1),
            'suspend_pull_down': (2, 2),
            'has_serial': (2, 3),
        }
        hwords = {
            'vendor_id': 0x02,
            'product_id': 0x04,
            'type': 0x06,
        }
        if self.device_version in (0x0400, 0x0500):
            # Type BM and 2232C/D use 0xc to encode the USB version to expose
            # H device use this location to encode bus/group properties
            hwords['usb_version'] = 0x0c
            confs['use_usb_version'] = (2, 4)
        if name in hwords:
            val = to_int(value)
            if not 0 <= val <= 0xFFFF:
                raise ValueError('Invalid value for %s' % name)
            offset = hwords[name]
            self._eeprom[offset:offset+2] = spack('<H', val)
            if self.is_mirroring_enabled:
                # duplicate in 'sector 2'
                offset2 = self.mirror_sector + offset
                self._eeprom[offset2:offset2+2] = spack('<H', val)
            self._dirty.add(name)
            return
        if name in confs:
            val = to_bool(value, permissive=False, allow_int=True)
            offset, bit = confs[name]
            mask = 1 << bit
            if val:
                idx = 0x08 + offset
                self._eeprom[idx] |= mask
                if self.is_mirroring_enabled:
                    # duplicate in 'sector 2'
                    idx2 = self.mirror_sector + idx
                    self._eeprom[idx2] |= mask
            else:
                idx = 0x0a + offset
                self._eeprom[idx] &= ~mask
                if self.is_mirroring_enabled:
                    # duplicate in 'sector 2'
                    idx2 = self.mirror_sector + idx
                    self._eeprom[idx2] &= ~mask
            self._dirty.add(name)
            return
        if name == 'power_max':
            val = to_int(value) >> 1
            idx = 0x09
            self._eeprom[idx] = val
            if self.is_mirroring_enabled:
                # duplicate in 'sector 2'
                idx2 = self.mirror_sector + idx
                self._eeprom[idx2] = val
            self._dirty.add(name)
            return
        if name.startswith('invert_'):
            if not self.device_version in (0x600, 0x1000):
                raise ValueError('UART control line inversion not available '
                                 'with this device')
            self._set_invert(name[len('invert_'):], value, out)
            self._dirty.add(name)
            return
        if name in self.properties:
            if name not in self._config:
                raise NotImplementedError("change is not supported")
            curval = self._config[name]
            try:
                curtype = type(curval)
                value = curtype(value)
            except (ValueError, TypeError) as exc:
                raise ValueError("cannot be converted to "
                    "the proper type '%s'" % curtype) from exc
            if value != curval:
                raise NotImplementedError("not yet supported")
            # no-op change is silently ignored
            return
        raise ValueError(f"unknown property: {name}")

    def erase(self, erase_byte: Optional[int] = 0xFF) -> None:
        """Erase the whole EEPROM.

            :param erase_byte: Optional erase byte to use. Default to 0xFF
        """
        self._eeprom = bytearray([erase_byte] * self.size)
        self._config.clear()
        self._dirty.add('eeprom')

    def initialize(self) -> None:
        """Initialize the EEPROM with some default sensible values.
        """
        dev_ver = self.device_version
        dev_name = Ftdi.DEVICE_NAMES[dev_ver]
        vid = Ftdi.FTDI_VENDOR
        pid = Ftdi.PRODUCT_IDS[vid][dev_name]
        self.set_manufacturer_name('FTDI')
        self.set_product_name(dev_name.upper())
        sernum = ''.join([chr(randint(ord('A'), ord('Z'))) for _ in range(5)])
        self.set_serial_number('FT%d%s' % (randint(0, 9), sernum))
        self.set_property('vendor_id', vid)
        self.set_property('product_id', pid)
        self.set_property('type', dev_ver)
        self.set_property('power_max', 150)
        self._sync_eeprom()

    def sync(self) -> None:
        """Force re-evaluation of configuration after some changes.

           This API is not useful for regular usage, but might help for testing
           when the EEPROM does not go through a full save/load cycle
        """
        self._sync_eeprom()

    def dump_config(self, file: Optional[BinaryIO] = None) -> None:
        """Dump the configuration to a file.

           :param file: the output file, default to stdout
        """
        if self._dirty:
            self._decode_eeprom()
        for name, value in self._config.items():
            print('%s: %s' % (name, value), file=file or sys.stdout)

    def commit(self, dry_run: bool = True, no_crc: bool = False) -> bool:
        """Commit any changes to the EEPROM.

           :param dry_run: log what should be written, do not actually change
                  the EEPROM content
           :param no_crc: do not compute EEPROM CRC. This should only be used
            to perform a full erasure of the EEPROM, as an attempt to recover
            from a corrupted config. 

           :return: True if some changes have been committed to the EEPROM
        """
        self._sync_eeprom(no_crc)
        if not self._modified:
            self.log.warning('No change to commit, EEPROM not modified')
            return False
        self._ftdi.overwrite_eeprom(self._eeprom, dry_run=dry_run)
        if not dry_run:
            eeprom = self._read_eeprom()
            if eeprom != self._eeprom:
                pos = 0
                for pos, (old, new) in enumerate(zip(self._eeprom, eeprom)):
                    if old != new:
                        break
                pos &= ~0x1
                raise FtdiEepromError('Write to EEPROM failed @ 0x%02x' % pos)
            self._modified = False
        return dry_run

    def reset_device(self):
        """Execute a USB device reset."""
        self._ftdi.reset(usb_reset=True)

    @classmethod
    def _validate_string(cls, string):
        for invchr in ':/':
            # do not accept characters which are interpreted as URL seperators
            if invchr in string:
                raise ValueError("Invalid character '%s' in string" % invchr)

    def _update_var_string(self, name: str, value: str) -> None:
        if name not in self.VAR_STRINGS:
            raise ValueError('%s is not a variable string' % name)
        try:
            if value == self._config[name]:
                return
        except KeyError:
            # not yet defined
            pass
        self._config[name] = value
        self._dirty.add(name)

    def _generate_var_strings(self, fill=True) -> None:
        """
            :param fill: fill the remainder of the space after the var strings
                with 0s
        """
        stream = bytearray()
        dynpos = self._PROPERTIES[self.device_version].dynoff
        if dynpos > self._size:
            # if a custom,small EEPROM device is used
            dynpos = 0x40
        data_pos = dynpos
        # start of var-strings in sector 1 (used for mirrored config)
        s1_vstr_start = data_pos - self.mirror_sector
        tbl_pos = 0x0e
        tbl_sector2_pos = self.mirror_sector + tbl_pos
        for name in self.VAR_STRINGS:
            try:
                ustr = self._config[name].encode('utf-16le')
            except KeyError:
                ustr = ''
            length = len(ustr)+2
            stream.append(length)
            stream.append(0x03)  # string descriptor
            stream.extend(ustr)
            self._eeprom[tbl_pos] = data_pos
            if self.is_mirroring_enabled:
                self._eeprom[tbl_sector2_pos] = data_pos
            tbl_pos += 1
            tbl_sector2_pos += 1
            self._eeprom[tbl_pos] = length
            if self.is_mirroring_enabled:
                self._eeprom[tbl_sector2_pos] = length
            tbl_pos += 1
            tbl_sector2_pos += 1
            data_pos += length
        if self.is_mirroring_enabled:
            self._eeprom[s1_vstr_start:s1_vstr_start+len(stream)] = stream
        self._eeprom[dynpos:dynpos+len(stream)] = stream
        mtp = self._ftdi.device_version == 0x1000
        crc_pos = 0x100 if mtp else self._size
        rem = crc_pos - (dynpos + len(stream))
        if rem < 0:
            oversize = (-rem + 2) // 2
            raise FtdiEepromError(f'Cannot fit strings into EEPROM, '
                                  f'{oversize} oversize characters')
        if fill:
            self._eeprom[dynpos+len(stream):crc_pos] = bytes(rem)
            if self.is_mirroring_enabled:
                crc_s1_pos = self.mirror_sector
                self._eeprom[s1_vstr_start+len(stream):crc_s1_pos] = bytes(rem)

    def _sync_eeprom(self, no_crc: bool = False):
        if not self._dirty:
            self.log.debug('No change detected for EEPROM content')
            return
        if not no_crc:
            if any([x in self._dirty for x in self.VAR_STRINGS]):
                self._generate_var_strings()
                for varstr in self.VAR_STRINGS:
                    self._dirty.discard(varstr)
            self._update_crc()
            self._decode_eeprom()
        self._dirty.clear()
        self._modified = True
        self.log.debug('EEPROM content regenerated (not yet committed)')

    def _compute_crc(self, eeprom: Union[bytes, bytearray], check=False):
        mtp = self._ftdi.device_version == 0x1000
        crc_pos = 0x100 if mtp else len(eeprom)
        crc_size = scalc('<H')
        if not check:
            # check mode: add CRC itself, so that result should be zero
            crc_pos -= crc_size
        if self.is_mirroring_enabled:
            mirror_s1_crc_pos = self.mirror_sector
            if not check:
                mirror_s1_crc_pos -= crc_size
            # if mirroring, only calculate the crc for the first sector/half
            #   of the eeprom. Data (including this crc) are duplicated in
            #   the second sector/half
            crc = self._ftdi.calc_eeprom_checksum(eeprom[:mirror_s1_crc_pos])
        else:
            crc = self._ftdi.calc_eeprom_checksum(eeprom[:crc_pos])
        if check:
            self._valid = not bool(crc)
            if not self._valid:
                self.log.debug('CRC is now 0x%04x', crc)
            else:
                self.log.debug('CRC OK')
        return crc, crc_pos, crc_size

    def _update_crc(self):
        crc, crc_pos, crc_size = self._compute_crc(
            self._eeprom, False)
        self._eeprom[crc_pos:crc_pos+crc_size] = spack('<H', crc)
        if self.is_mirroring_enabled:
            # if mirroring calculate where the CRC will start in first sector
            crc_s1_start = self.mirror_sector - crc_size
            self._eeprom[crc_s1_start:crc_s1_start+crc_size] = spack('<H', crc)

    def _compute_size(self, 
            eeprom: Union[bytes, bytearray]) -> Tuple[int, bool]:
        """
            :return: Tuple of:
                - int of usable size of the eeprom
                - bool of whether eeprom mirroring was detected or not
        """
        if self._ftdi.is_eeprom_internal:
            return self._ftdi.max_eeprom_size, False
        if all([x == 0xFF for x in eeprom]):
            # erased EEPROM, size is unknown
            return self._ftdi.max_eeprom_size, False
        if eeprom[0:0x80] == eeprom[0x80:0x100]:
            return 0x80, True
        if eeprom[0:0x40] == eeprom[0x40:0x80]:
            return 0x40, True
        return 0x100, False

    def _read_eeprom(self) -> bytes:
        buf = self._ftdi.read_eeprom(0, eeprom_size=self.size)
        eeprom = bytearray(buf)
        size, mirror_detected = self._compute_size(eeprom)
        if size < len(eeprom):
            eeprom = eeprom[:size]
        crc = self._compute_crc(eeprom, True)[0]
        if crc:
            if self.is_empty:
                self.log.info('No EEPROM or EEPROM erased')
            else:
                self.log.error('Invalid CRC or EEPROM content')
        if not self.is_empty and mirror_detected:
            self.log.info("Detected a mirrored eeprom. " +
                "Enabling mirrored writing")
            self._mirror = True
        return eeprom

    def _decode_eeprom(self):
        cfg = self._config
        cfg.clear()
        chipoff = self._PROPERTIES[self.device_version].chipoff
        if chipoff is not None:
            self._chip = Hex2Int(self._eeprom[chipoff])
            cfg['chip'] = self._chip
        cfg['vendor_id'] = Hex4Int(sunpack('<H', self._eeprom[0x02:0x04])[0])
        cfg['product_id'] = Hex4Int(sunpack('<H', self._eeprom[0x04:0x06])[0])
        cfg['type'] = Hex4Int(sunpack('<H', self._eeprom[0x06:0x08])[0])
        power_supply, power_max, conf = sunpack('<3B', self._eeprom[0x08:0x0b])
        cfg['self_powered'] = bool(power_supply & (1 << 6))
        cfg['remote_wakeup'] = bool(power_supply & (1 << 5))
        cfg['power_max'] = power_max << 1
        cfg['has_serial'] = bool(conf & (1 << 3))
        cfg['suspend_pull_down'] = bool(conf & (1 << 2))
        cfg['out_isochronous'] = bool(conf & (1 << 1))
        cfg['in_isochronous'] = bool(conf & (1 << 0))
        cfg['manufacturer'] = self._decode_string(0x0e)
        cfg['product'] = self._decode_string(0x10)
        cfg['serial'] = self._decode_string(0x12)
        if self.device_version in (0x0400, 0x0500):
            cfg['use_usb_version'] = bool(conf & (1 << 3))
            if cfg['use_usb_version']:
                cfg['usb_version'] = \
                    Hex4Int(sunpack('<H', self._eeprom[0x0c:0x0e])[0])
        name = None
        try:
            name = Ftdi.DEVICE_NAMES[cfg['type']].replace('-', '')
            if name.startswith('ft'):
                name = name[2:]
            func = getattr(self, '_decode_%s' % name)
        except (KeyError, AttributeError):
            self.log.warning('No EEPROM decoder for device %s', name or '?')
        else:
            func()

    def _decode_string(self, offset):
        str_offset, str_size = sunpack('<BB', self._eeprom[offset:offset+2])
        if str_size:
            str_offset &= self.size - 1
            str_size -= scalc('<H')
            str_offset += scalc('<H')
            manufacturer = self._eeprom[str_offset:str_offset+str_size]
            return manufacturer.decode('utf16', errors='ignore')
        return ''

    def _set_cbus_func(self, cpin: int, value: str,
                       out: Optional[TextIO]) -> None:
        cmap = {0x600: (self.CBUS, 5, 0x14, 4),    # FT232R
                0x900: (self.CBUSH, 10, 0x18, 4),  # FT232H
                0x1000: (self.CBUSX, 4, 0x1A, 8)}  # FT230X/FT231X/FT234X
        try:
            cbus, count, offset, width = cmap[self.device_version]
        except KeyError as exc:
            raise ValueError('This property is not supported on this '
                             'device') from exc
        pin_filter = getattr(self,
                             '_filter_cbus_func_x%x' % self.device_version,
                             None)
        if value == '?' and out:
            items = {item.name for item in cbus}
            if pin_filter:
                items = {val for val in items if pin_filter(cpin, val)}
            print(', '.join(sorted(items)) if items else '(none)', file=out)
            return
        if not 0 <= cpin < count:
            raise ValueError("Unsupported CBUS pin '%d'" % cpin)
        try:
            code = cbus[value.upper()].value
        except KeyError as exc:
            raise ValueError("CBUS pin %d does not have function '%s'" %
                             (cpin, value)) from exc
        if pin_filter and not pin_filter(cpin, value.upper()):
            raise ValueError("Unsupported CBUS function '%s' for pin '%d'" %
                             (value, cpin))
        addr = offset + (cpin*width)//8
        if width == 4:
            bitoff = 4 if cpin & 0x1 else 0
            mask = 0x0F << bitoff
        else:
            bitoff = 0
            mask = 0xFF
        old = self._eeprom[addr]
        self._eeprom[addr] &= ~mask
        self._eeprom[addr] |= code << bitoff
        self.log.debug('Cpin %d, addr 0x%02x, value 0x%02x->0x%02x',
                       cpin, addr, old, self._eeprom[addr])

    @classmethod
    def _filter_cbus_func_x900(cls, cpin: int, value: str):
        if cpin == 7:
            # nothing can be assigned to ACBUS7
            return False
        if value in 'TRISTATE TXLED RXLED TXRXLED PWREN SLEEP DRIVE0'.split():
            # any pin can be assigned these functions
            return True
        if cpin in (5, 6, 8, 9):
            # any function can be assigned to ACBUS5, ACBUS6, ACBUS8, ACBUS9
            return True
        if cpin == 0:
            return value != 'GPIO'
        return False

    @classmethod
    def _filter_cbus_func_x600(cls, cpin: int, value: str):
        if value == 'BB_WR':
            # this signal is only available on CBUS0, CBUS1
            return cpin < 2
        return True

    def _set_bus_control(self, bus: str, control: str,
                         value: Union[str, int, bool],
                         out: Optional[TextIO]) -> None:
        if self.device_version == 0x1000:
            self._set_bus_control_230x(bus, control, value, out)
            return
        # for now, only support FT-X devices
        raise ValueError('Bus control not implemented for this device')

    def _set_group(self, group: int, control: str,
                   value: Union[str, int, bool], out: Optional[TextIO]) \
            -> None:
        if self.device_version in (0x0700, 0x0800, 0x0900):
            self._set_group_x232h(group, control, value, out)
            return
        raise ValueError('Group not implemented for this device')

    def _set_bus_control_230x(self, bus: str, control: str,
                              value: Union[str, int, bool],
                              out: Optional[TextIO]) -> None:
        if bus not in 'cd':
            raise ValueError('Invalid bus: %s' % bus)
        self._set_bus_xprop(0x0c, bus == 'c', control, value, out)

    def _set_group_x232h(self, group: int, control: str, value: str,
                         out: Optional[TextIO]) -> None:
        if self.device_version in (0x0700, 0x800):  # 2232H/4232H
            offset = 0x0c + group//2
            nibble = group & 1
        else:  # 232H
            offset = 0x0c + group
            nibble = 0
        self._set_bus_xprop(offset, nibble, control, value, out)

    def _set_bus_xprop(self, offset: int, high_nibble: bool, control: str,
                       value: Union[str, int, bool], out: Optional[TextIO]) \
            -> None:
        try:
            if control == 'drive':
                candidates = (4, 8, 12, 16)
                if value == '?' and out:
                    print(', '.join([str(v) for v in candidates]), file=out)
                    return
                value = int(value)
                if value not in candidates:
                    raise ValueError('Invalid drive current: %d mA' % value)
                value //= 4
                value -= 1
            elif control in ('slow_slew', 'schmitt'):
                if value == '?' and out:
                    print('off, on', file=out)
                    return
                value = int(to_bool(value))
            else:
                raise ValueError('Unsupported control: %s' % control)
        except (ValueError, TypeError) as exc:
            raise ValueError('Invalid %s value: %s' %
                             (control, value)) from exc
        config = self._eeprom[offset]
        if not high_nibble:
            conf = config & 0x0F
            config &= 0xF0
            cshift = 0
        else:
            conf = config >> 4
            config &= 0x0F
            cshift = 4
        if control == 'drive':
            conf &= 0b1100
            conf |= value
        elif control == 'slow_slew':
            conf &= 0b1011
            conf |= value << 2
        elif control == 'schmitt':
            conf &= 0b0111
            conf |= value << 3
        else:
            raise RuntimeError('Internal error')
        config |= conf << cshift
        self._eeprom[offset] = config

    def _set_invert(self, name, value, out):
        if value == '?' and out:
            print('off, on', file=out)
            return
        if name.upper() not in self.UART_BITS.__members__:
            raise ValueError('Unknown property: %s' % name)
        value = to_bool(value, permissive=False)
        code = getattr(self.UART_BITS, name.upper())
        invert = self._eeprom[0x0B]
        if value:
            invert |= code
        else:
            invert &= ~code
        self._eeprom[0x0B] = invert

    def _decode_x(self):
        # FT-X series
        cfg = self._config
        misc, = sunpack('<H', self._eeprom[0x00:0x02])
        cfg['channel_a_driver'] = 'VCP' if misc & (1 << 7) else 'D2XX'
        for bit in self.UART_BITS:
            value = self._eeprom[0x0B]
            cfg['invert_%s' % self.UART_BITS(bit).name] = bool(value & bit)
        max_drive = self.DRIVE.LOW.value | self.DRIVE.HIGH.value
        value = self._eeprom[0x0c]
        for grp in range(2):
            conf = value & 0xF
            bus = 'c' if grp else 'd'
            cfg['%sbus_drive' % bus] = 4 * (1+(conf & max_drive))
            cfg['%sbus_schmitt' % bus] = bool(conf & self.DRIVE.SCHMITT)
            cfg['%sbus_slow_slew' % bus] = bool(conf & self.DRIVE.SLOW_SLEW)
            value >>= 4
        for bix in range(4):
            value = self._eeprom[0x1A + bix]
            try:
                cfg['cbus_func_%d' % bix] = self.CBUSX(value).name
            except ValueError:
                pass

    def _decode_232h(self):
        cfg = self._config
        cfg0, cfg1 = self._eeprom[0x00], self._eeprom[0x01]
        cfg['channel_a_type'] = cfg0 & 0x0F
        cfg['channel_a_driver'] = 'VCP' if (cfg0 & (1 << 4)) else 'D2XX'
        cfg['clock_polarity'] = 'high' if (cfg1 & self.CFG1.CLK_IDLE_STATE) \
                                else 'low'
        cfg['lsb_data'] = bool(cfg1 & self.CFG1.DATA_LSB)
        cfg['flow_control'] = 'on' if (cfg1 & self.CFG1.FLOW_CONTROL) \
                              else 'off'
        cfg['powersave'] = bool(cfg1 & self.DRIVE.PWRSAVE_DIS)
        max_drive = self.DRIVE.LOW.value | self.DRIVE.HIGH.value
        for grp in range(2):
            conf = self._eeprom[0x0c+grp]
            cfg['group_%d_drive' % grp] = 4 * (1+(conf & max_drive))
            cfg['group_%d_schmitt' % grp] = \
                bool(conf & self.DRIVE.SCHMITT.value)
            cfg['group_%d_slow_slew' % grp] = \
                bool(conf & self.DRIVE.SLOW_SLEW.value)
        for bix in range(5):
            value = self._eeprom[0x18 + bix]
            low, high = value & 0x0F, value >> 4
            try:
                cfg['cbus_func_%d' % ((2*bix)+0)] = self.CBUSH(low).name
            except ValueError:
                pass
            try:
                cfg['cbus_func_%d' % ((2*bix)+1)] = self.CBUSH(high).name
            except ValueError:
                pass

    def _decode_232r(self):
        cfg = self._config
        cfg0 = self._eeprom[0x00]
        cfg['channel_a_driver'] = 'VCP' if (~cfg0 & (1 << 3)) else ''
        cfg['high_current'] = bool(~cfg0 & (1 << 2))
        cfg['external_oscillator'] = cfg0 & 0x02
        for bit in self.UART_BITS:
            value = self._eeprom[0x0B]
            cfg['invert_%s' % self.UART_BITS(bit).name] = bool(value & bit)
        bix = 0
        while True:
            value = self._eeprom[0x14 + bix]
            low, high = value & 0x0F, value >> 4
            try:
                cfg['cbus_func_%d' % ((2*bix)+0)] = self.CBUS(low).name
            except ValueError:
                pass
            if bix == 2:
                break
            try:
                cfg['cbus_func_%d' % ((2*bix)+1)] = self.CBUS(high).name
            except ValueError:
                pass
            bix += 1

    def _decode_2232h(self):
        cfg = self._config
        self._decode_x232h(cfg)
        cfg0, cfg1 = self._eeprom[0x00], self._eeprom[0x01]
        cfg['channel_a_type'] = self.CHANNEL(cfg0 & 0x7).name or 'UART'
        cfg['channel_b_type'] = self.CHANNEL(cfg1 & 0x7).name or 'UART'
        cfg['suspend_dbus7'] = bool(cfg1 & self.CFG1.SUSPEND_DBUS7.value)

    def _decode_4232h(self):
        cfg = self._config
        self._decode_x232h(cfg)
        cfg0, cfg1 = self._eeprom[0x00], self._eeprom[0x01]
        cfg['channel_c_driver'] = 'VCP' if ((cfg0 >> 4) & (1 << 3)) else 'D2XX'
        cfg['channel_d_driver'] = 'VCP' if ((cfg1 >> 4) & (1 << 3)) else 'D2XX'
        conf = self._eeprom[0x0B]
        rs485 = self.CHANNEL.RS485
        for chix in range(4):
            cfg['channel_%x_rs485' % (0xa+chix)] = bool(conf & (rs485 << chix))

    def _decode_x232h(self, cfg):
        # common code for 2232h and 4232h
        cfg0, cfg1 = self._eeprom[0x00], self._eeprom[0x01]
        cfg['channel_a_driver'] = 'VCP' if (cfg0 & (1 << 3)) else 'D2XX'
        cfg['channel_b_driver'] = 'VCP' if (cfg1 & (1 << 3)) else 'D2XX'
        max_drive = self.DRIVE.LOW.value | self.DRIVE.HIGH.value
        for bix in range(4):
            if not bix & 1:
                val = self._eeprom[0x0c + bix//2]
            else:
                val >>= 4
            cfg['group_%d_drive' % bix] = 4 * (1+(val & max_drive))
            cfg['group_%d_schmitt' % bix] = \
                bool(val & self.DRIVE.SCHMITT.value)
            cfg['group_%d_slow_slew' % bix] = \
                bool(val & self.DRIVE.SLOW_SLEW.value)



# Copyright (c) 2017-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""MPSSE command debug tracer."""

#pylint: disable-msg=missing-docstring
#pylint: disable-msg=too-many-instance-attributes

from binascii import hexlify
from collections import deque
from inspect import currentframe
from logging import getLogger
from string import ascii_uppercase
from struct import unpack as sunpack
from typing import Union
from .ftdi import Ftdi


class FtdiMpsseTracer:
    """FTDI MPSSE protocol decoder."""

    MPSSE_ENGINES = {
        0x0200: 0,
        0x0400: 0,
        0x0500: 0,
        0x0600: 0,
        0x0700: 2,
        0x0800: 2,
        0x0900: 1,
        0x1000: 0}
    """Count of MPSSE engines."""

    def __init__(self, version):
        count = self.MPSSE_ENGINES[version]
        self._engines = [None] * count

    def send(self, iface: int, buf: Union[bytes, bytearray]) -> None:
        self._get_engine(iface).send(buf)

    def receive(self, iface: int, buf: Union[bytes, bytearray]) -> None:
        self._get_engine(iface).receive(buf)

    def _get_engine(self, iface: int):
        iface -= 1
        try:
            self._engines[iface]
        except IndexError as exc:
            raise ValueError('No MPSSE engine available on interface %d' %
                             iface) from exc
        if not self._engines[iface]:
            self._engines[iface] = FtdiMpsseEngine(iface)
        return self._engines[iface]


class FtdiMpsseEngine:
    """FTDI MPSSE virtual engine

       Far from being complete for now
    """

    COMMAND_PREFIX = \
        'GET SET READ WRITE RW ENABLE DISABLE CLK LOOPBACK SEND DRIVE'

    def build_commands(prefix: str):
        commands = {}
        for cmd in dir(Ftdi):
            if cmd[0] not in ascii_uppercase:
                continue
            value = getattr(Ftdi, cmd)
            if not isinstance(value, int):
                continue
            family = cmd.split('_')[0]
            if family not in prefix.split():
                continue
            commands[value] = cmd
        return commands

    COMMANDS = build_commands(COMMAND_PREFIX)

    ST_IDLE = range(1)

    def __init__(self, iface: int):
        self.log = getLogger('pyftdi.mpsse.tracer')
        self._if = iface
        self._trace_tx = bytearray()
        self._trace_rx = bytearray()
        self._state = self.ST_IDLE
        self._clkdiv5 = False
        self._cmd_decoded = True
        self._resp_decoded = True
        self._last_codes = deque()
        self._expect_resp = deque()  # positive: byte, negative: bit count

    def send(self, buf: Union[bytes, bytearray]) -> None:
        self._trace_tx.extend(buf)
        while self._trace_tx:
            try:
                code = self._trace_tx[0]
                cmd = self.COMMANDS[code]
                if self._cmd_decoded:
                    self.log.debug('[%d]:[Command: %02X: %s]',
                                   self._if, code, cmd)
                cmd_decoder = getattr(self, '_cmd_%s' % cmd.lower())
                rdepth = len(self._expect_resp)
                try:
                    self._cmd_decoded = cmd_decoder()
                except AttributeError as exc:
                    raise ValueError(str(exc)) from exc
                if len(self._expect_resp) > rdepth:
                    self._last_codes.append(code)
                if self._cmd_decoded:
                    continue
                # not enough data in buffer to decode a whole command
                return
            except IndexError:
                self.log.warning('[%d]:Empty buffer on %02X: %s',
                                 self._if, code, cmd)
            except KeyError:
                self.log.warning('[%d]:Unknown command code: %02X',
                                 self._if, code)
            except AttributeError as exc:
                self.log.warning('[%d]:Decoder for command %s [%02X] is not '
                                 'implemented', self._if, cmd, code)
            except ValueError as exc:
                self.log.warning('[%d]:Decoder for command %s [%02X] failed: '
                                 '%s', self._if, cmd, code, exc)
            # on error, flush all buffers
            self.log.warning('Flush TX/RX buffers')
            self._trace_tx = bytearray()
            self._trace_rx = bytearray()
            self._last_codes.clear()

    def receive(self, buf: Union[bytes, bytearray]) -> None:
        self.log.info(' .. %s', hexlify(buf).decode())
        self._trace_rx.extend(buf)
        while self._trace_rx:
            code = None
            try:
                code = self._last_codes.popleft()
                cmd = self.COMMANDS[code]
                resp_decoder = getattr(self, '_resp_%s' % cmd.lower())
                self._resp_decoded = resp_decoder()
                if self._resp_decoded:
                    continue
                # not enough data in buffer to decode a whole response
                return
            except IndexError:
                self.log.warning('[%d]:Empty buffer', self._if)
            except KeyError:
                self.log.warning('[%d]:Unknown command code: %02X',
                                 self._if, code)
            except AttributeError:
                self.log.warning('[%d]:Decoder for response %s [%02X] is not '
                                 'implemented', self._if, cmd, code)
            # on error, flush RX buffer
            self.log.warning('[%d]:Flush RX buffer', self._if)
            self._trace_rx = bytearray()
            self._last_codes.clear()

    def _cmd_enable_clk_div5(self):
        self.log.info(' [%d]:Enable clock divisor /5', self._if)
        self._clkdiv5 = True
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_disable_clk_div5(self):
        self.log.info(' [%d]:Disable clock divisor /5', self._if)
        self._clkdiv5 = False
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_set_tck_divisor(self):
        if len(self._trace_tx) < 3:
            return False
        value, = sunpack('<H', self._trace_tx[1:3])
        base = 12E6 if self._clkdiv5 else 60E6
        freq = base / ((1 + value) * 2)
        self.log.info(' [%d]:Set frequency %.3fMHZ', self._if, freq/1E6)
        self._trace_tx[:] = self._trace_tx[3:]
        return True

    def _cmd_loopback_end(self):
        self.log.info(' [%d]:Disable loopback', self._if)
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_enable_clk_adaptive(self):
        self.log.info(' [%d]:Enable adaptive clock', self._if)
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_disable_clk_adaptive(self):
        self.log.info(' [%d]:Disable adaptive clock', self._if)
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_enable_clk_3phase(self):
        self.log.info(' [%d]:Enable 3-phase clock', self._if)
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_disable_clk_3phase(self):
        self.log.info(' [%d]:Disable 3-phase clock', self._if)
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_drive_zero(self):
        if len(self._trace_tx) < 3:
            return False
        value, = sunpack('H', self._trace_tx[1:3])
        self.log.info(' [%d]:Open collector [15:0] %04x %s',
                      self._if, value, self.bitfmt(value, 16))
        self._trace_tx[:] = self._trace_tx[3:]
        return True

    def _cmd_send_immediate(self):
        self.log.debug(' [%d]:Send immediate', self._if)
        self._trace_tx[:] = self._trace_tx[1:]
        return True

    def _cmd_get_bits_low(self):
        self._trace_tx[:] = self._trace_tx[1:]
        self._expect_resp.append(1)
        return True

    def _cmd_get_bits_high(self):
        self._trace_tx[:] = self._trace_tx[1:]
        self._expect_resp.append(1)
        return True

    def _cmd_set_bits_low(self):
        if len(self._trace_tx) < 3:
            return False
        value, direction = sunpack('BB', self._trace_tx[1:3])
        self.log.info(' [%d]:Set gpio[7:0]  %02x %s',
                      self._if, value, self.bm2str(value, direction))
        self._trace_tx[:] = self._trace_tx[3:]
        return True

    def _cmd_set_bits_high(self):
        if len(self._trace_tx) < 3:
            return False
        value, direction = sunpack('BB', self._trace_tx[1:3])
        self.log.info(' [%d]:Set gpio[15:8] %02x %s',
                      self._if, value, self.bm2str(value, direction))
        self._trace_tx[:] = self._trace_tx[3:]
        return True

    def _cmd_write_bytes_pve_msb(self):
        return self._decode_output_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_write_bytes_nve_msb(self):
        return self._decode_output_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_write_bytes_pve_lsb(self):
        return self._decode_output_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_write_bytes_nve_lsb(self):
        return self._decode_output_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_read_bytes_pve_msb(self):
        return self._decode_input_mpsse_byte_request()

    def _resp_read_bytes_pve_msb(self):
        return self._decode_input_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_read_bytes_nve_msb(self):
        return self._decode_input_mpsse_byte_request()

    def _resp_read_bytes_nve_msb(self):
        return self._decode_input_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_read_bytes_pve_lsb(self):
        return self._decode_input_mpsse_byte_request()

    def _resp_read_bytes_pve_lsb(self):
        return self._decode_input_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_read_bytes_nve_lsb(self):
        return self._decode_input_mpsse_byte_request()

    def _resp_read_bytes_nve_lsb(self):
        return self._decode_input_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_rw_bytes_nve_pve_msb(self):
        return self._decode_output_mpsse_bytes(currentframe().f_code.co_name,
                                               True)

    def _resp_rw_bytes_nve_pve_msb(self):
        return self._decode_input_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_rw_bytes_pve_nve_msb(self):
        return self._decode_output_mpsse_bytes(currentframe().f_code.co_name,
                                               True)

    def _resp_rw_bytes_pve_nve_msb(self):
        return self._decode_input_mpsse_bytes(currentframe().f_code.co_name)

    def _cmd_write_bits_pve_msb(self):
        return self._decode_output_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_write_bits_nve_msb(self):
        return self._decode_output_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_write_bits_pve_lsb(self):
        return self._decode_output_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_write_bits_nve_lsb(self):
        return self._decode_output_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_read_bits_pve_msb(self):
        return self._decode_input_mpsse_bit_request()

    def _resp_read_bits_pve_msb(self):
        return self._decode_input_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_read_bits_nve_msb(self):
        return self._decode_input_mpsse_bit_request()

    def _resp_read_bits_nve_msb(self):
        return self._decode_input_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_read_bits_pve_lsb(self):
        return self._decode_input_mpsse_bit_request()

    def _resp_read_bits_pve_lsb(self):
        return self._decode_input_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_read_bits_nve_lsb(self):
        return self._decode_input_mpsse_bit_request()

    def _resp_read_bits_nve_lsb(self):
        return self._decode_input_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_rw_bits_nve_pve_msb(self):
        return self._decode_output_mpsse_bits(currentframe().f_code.co_name,
                                              True)

    def _resp_rw_bits_nve_pve_msb(self):
        return self._decode_input_mpsse_bits(currentframe().f_code.co_name)

    def _cmd_rw_bits_pve_nve_msb(self):
        return self._decode_output_mpsse_bits(currentframe().f_code.co_name,
                                              True)

    def _resp_rw_bits_pve_nve_msb(self):
        return self._decode_input_mpsse_bits(currentframe().f_code.co_name)

    def _resp_get_bits_low(self):
        if self._trace_rx:
            return False
        value = self._trace_rx[0]
        self.log.info(' [%d]:Get gpio[7:0]  %02x %s',
                      self._if, value, self.bm2str(value, 0xFF))
        self._trace_rx[:] = self._trace_rx[1:]
        return True

    def _resp_get_bits_high(self):
        if self._trace_rx:
            return False
        value = self._trace_rx[0]
        self.log.info(' [%d]:Get gpio[15:8] %02x %s',
                      self._if, value, self.bm2str(value, 0xFF))
        self._trace_rx[:] = self._trace_rx[1:]
        return True

    def _decode_output_mpsse_bytes(self, caller, expect_rx=False):
        if len(self._trace_tx) < 4:
            return False
        length = sunpack('<H', self._trace_tx[1:3])[0] + 1
        if len(self._trace_tx) < 4 + length:
            return False
        if expect_rx:
            self._expect_resp.append(length)
        payload = self._trace_tx[3:3+length]
        funcname = caller[5:].title().replace('_', '')
        self.log.info(' [%d]:%s> (%d) %s',
                      self._if, funcname, length,
                      hexlify(payload).decode('utf8'))
        self._trace_tx[:] = self._trace_tx[3+length:]
        return True

    def _decode_output_mpsse_bits(self, caller, expect_rx=False):
        if len(self._trace_tx) < 3:
            return False
        bitlen = self._trace_tx[1] + 1
        if expect_rx:
            self._expect_resp.append(-bitlen)
        payload = self._trace_tx[2]
        funcname = caller[5:].title().replace('_', '')
        msb = caller[5:][-3].lower() == 'm'
        self.log.info(' %s> (%d) %s',
                      funcname, bitlen, self.bit2str(payload, bitlen, msb))
        self._trace_tx[:] = self._trace_tx[3:]
        return True

    def _decode_input_mpsse_byte_request(self):
        if len(self._trace_tx) < 3:
            return False
        length = sunpack('<H', self._trace_tx[1:3])[0] + 1
        self._expect_resp.append(length)
        self._trace_tx[:] = self._trace_tx[3:]
        return True

    def _decode_input_mpsse_bit_request(self):
        if len(self._trace_tx) < 2:
            return False
        bitlen = self._trace_tx[1] + 1
        self._expect_resp.append(-bitlen)
        self._trace_tx[:] = self._trace_tx[2:]
        return True

    def _decode_input_mpsse_bytes(self, caller):
        if not self._expect_resp:
            self.log.warning('[%d]:Response w/o request?', self._if)
            return False
        if self._expect_resp[0] < 0:
            self.log.warning('[%d]:Handling byte request w/ bit length',
                             self._if)
            return False
        if len(self._trace_rx) < self._expect_resp[0]:  # peek
            return False
        length = self._expect_resp.popleft()
        payload = self._trace_rx[:length]
        self._trace_rx[:] = self._trace_rx[length:]
        funcname = caller[5:].title().replace('_', '')
        self.log.info(' %s< (%d) %s',
                      funcname, length, hexlify(payload).decode('utf8'))
        return True

    def _decode_input_mpsse_bits(self, caller):
        if not self._expect_resp:
            self.log.warning('[%d]:Response w/o request?', self._if)
            return False
        if not self._trace_rx:  # peek
            return False
        if self._expect_resp[0] > 0:
            self.log.warning('[%d]:Handling bit request w/ byte length',
                             self._if)
        bitlen = -self._expect_resp.popleft()
        payload = self._trace_rx[0]
        self._trace_rx[:] = self._trace_rx[1:]
        funcname = caller[5:].title().replace('_', '')
        msb = caller[5:][-3].lower() == 'm'
        self.log.info(' %s< (%d) %s',
                      funcname, bitlen, self.bit2str(payload, bitlen, msb))
        return True

    @classmethod
    def bit2str(cls, value: int, count: int, msb: bool, hiz: str = '_') -> str:
        mask = (1 << count) - 1
        if msb:
            mask <<= 8 - count
        return cls.bm2str(value, mask, hiz)

    @classmethod
    def bm2str(cls, value: int, mask: int, hiz: str = '_') -> str:
        vstr = cls.bitfmt(value, 8)
        mstr = cls.bitfmt(mask, 8)
        return ''.join([m == '1' and v or hiz for v, m in zip(vstr, mstr)])

    @classmethod
    def bitfmt(cls, value, width):
        return format(value, '0%db' % width)

    # rw_bytes_pve_pve_lsb
    # rw_bytes_pve_nve_lsb
    # rw_bytes_nve_pve_lsb
    # rw_bytes_nve_nve_lsb
    # rw_bits_pve_pve_lsb
    # rw_bits_pve_nve_lsb
    # rw_bits_nve_pve_lsb
    # rw_bits_nve_nve_lsb
    # write_bits_tms_pve
    # write_bits_tms_nve
    # rw_bits_tms_pve_pve
    # rw_bits_tms_nve_pve
    # rw_bits_tms_pve_nve
    # rw_bits_tms_nve_nve



# Copyright (c) 2014-2020, Emmanuel Blot <emmanuel.blot@free.fr>
# Copyright (c) 2016, Emmanuel Bouaziz <ebouaziz@free.fr>
# All rights reserved.
#
# SPDX-License-Identifier: BSD-3-Clause

"""GPIO/BitBang support for PyFdti"""

#pylint: disable-msg=too-few-public-methods

from struct import calcsize as scalc, unpack as sunpack
from typing import Iterable, Optional, Tuple, Union
from .ftdi import Ftdi, FtdiError
from .misc import is_iterable


class GpioException(FtdiError):
    """Base class for GPIO errors.
    """


class GpioPort:
    """Duck-type GPIO port for GPIO all controllers.
    """


class GpioBaseController(GpioPort):
    """GPIO controller for an FTDI port, in bit-bang legacy mode.

       GPIO bit-bang mode is limited to the 8 lower pins of each GPIO port.
    """

    def __init__(self):
        self._ftdi = Ftdi()
        self._direction = 0
        self._width = 0
        self._mask = 0
        self._frequency = 0

    @property
    def ftdi(self) -> Ftdi:
        """Return the Ftdi instance.

           :return: the Ftdi instance
        """
        return self._ftdi

    @property
    def is_connected(self) -> bool:
        """Reports whether a connection exists with the FTDI interface.

           :return: the FTDI slave connection status
        """
        return self._ftdi.is_connected

    def configure(self, url: str, direction: int = 0,
                  **kwargs) -> int:
        """Open a new interface to the specified FTDI device in bitbang mode.

           :param str url: a FTDI URL selector
           :param int direction: a bitfield specifying the FTDI GPIO direction,
                where high level defines an output, and low level defines an
                input
           :param initial: optional initial GPIO output value
           :param pace: optional pace in GPIO sample per second
           :return: actual bitbang pace in sample per second
        """
        if self.is_connected:
            raise FtdiError('Already connected')
        kwargs = dict(kwargs)
        frequency = kwargs.get('frequency', None)
        if frequency is None:
            frequency = kwargs.get('baudrate', None)
        for k in ('direction', 'sync', 'frequency', 'baudrate'):
            if k in kwargs:
                del kwargs[k]
        self._frequency = self._configure(url, direction, frequency, **kwargs)

    def close(self, freeze: bool = False) -> None:
        """Close the GPIO port.

           :param freeze: if set, FTDI port is not reset to its default
                          state on close. This means the port is left with
                          its current configuration and output signals.
                          This feature should not be used except for very
                          specific needs.
        """
        if self._ftdi.is_connected:
            self._ftdi.close(freeze)

    def get_gpio(self) -> GpioPort:
        """Retrieve the GPIO port.

           This method is mostly useless, it is a wrapper to duck type other
           GPIO APIs (I2C, SPI, ...)

           :return: GPIO port
        """
        return self

    @property
    def direction(self) -> int:
        """Reports the GPIO direction.

          :return: a bitfield specifying the FTDI GPIO direction, where high
                level reports an output pin, and low level reports an input pin
        """
        return self._direction

    @property
    def pins(self) -> int:
        """Report the configured GPIOs as a bitfield.

           A true bit represents a GPIO, a false bit a reserved or not
           configured pin.

           :return: always 0xFF for GpioController instance.
        """
        return self._mask

    @property
    def all_pins(self) -> int:
        """Report the addressable GPIOs as a bitfield.

           A true bit represents a pin which may be used as a GPIO, a false bit
           a reserved pin

           :return: always 0xFF for GpioController instance.
        """
        return self._mask

    @property
    def width(self) -> int:
        """Report the FTDI count of addressable pins.

           :return: the width of the GPIO port.
        """
        return self._width

    @property
    def frequency(self) -> float:
        """Return the pace at which sequence of GPIO samples are read
           and written.
        """
        return self._frequency

    def set_frequency(self, frequency: Union[int, float]) -> None:
        """Set the frequency at which sequence of GPIO samples are read
           and written.

           :param frequency: the new frequency, in GPIO samples per second
        """
        raise NotImplementedError('GpioBaseController cannot be instanciated')

    def set_direction(self, pins: int, direction: int) -> None:
        """Update the GPIO pin direction.

           :param pins: which GPIO pins should be reconfigured
           :param direction: a bitfield of GPIO pins. Each bit represent a
                GPIO pin, where a high level sets the pin as output and a low
                level sets the pin as input/high-Z.
        """
        if direction > self._mask:
            raise GpioException("Invalid direction mask")
        self._direction &= ~pins
        self._direction |= (pins & direction)
        self._update_direction()

    def _configure(self, url: str, direction: int,
                   frequency: Union[int, float, None] = None, **kwargs) -> int:
        raise NotImplementedError('GpioBaseController cannot be instanciated')

    def _update_direction(self) -> None:
        raise NotImplementedError('Missing implementation')


class GpioAsyncController(GpioBaseController):
    """GPIO controller for an FTDI port, in bit-bang asynchronous mode.

       GPIO accessible pins are limited to the 8 lower pins of each GPIO port.

       Asynchronous bitbang output are updated on write request using the
       :py:meth:`write` method, clocked at the selected frequency.

       Asynchronous bitbang input are sampled at the same rate, as soon as the
       controller is initialized. The GPIO input samples fill in the FTDI HW
       buffer until it is filled up, in which case sampling stops until the
       GPIO samples are read out with the :py:meth:`read` method. It may be
       therefore hard to use, except if peek mode is selected,
       see :py:meth:`read` for details.

       Note that FTDI internal clock divider cannot generate any arbitrary
       frequency, so the closest frequency to the request one that can be
       generated is selected. The actual :py:attr:`frequency` may be tested to
       check if it matches the board requirements.
    """

    def read(self, readlen: int = 1, peek: Optional[bool] = None,
             noflush: bool = False) -> Union[int, bytes]:
        """Read the GPIO input pin electrical level.

           :param readlen: how many GPIO samples to retrieve. Each sample is
                           8-bit wide.
           :param peek: whether to peek/sample the instantaneous GPIO pin
                        values from port, or to use the HW FIFO. The HW FIFO is
                        continously filled up with GPIO sample at the current
                        frequency, until it is full - samples are no longer
                        collected until the FIFO is read. This means than
                        non-peek mode read "old" values, with no way to know at
                        which time they have been sampled. PyFtdi ensures that
                        old sampled values before the completion of a previous
                        GPIO write are discarded. When peek mode is selected,
                        readlen should be 1.
           :param noflush: whether to disable the RX buffer flush before
                           reading out data
           :return: a 8-bit wide integer if peek mode is used, or
                    a bytes buffer otherwise.
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        if peek is None and readlen == 1:
            # compatibility with legacy API
            peek = True
        if peek:
            if readlen != 1:
                raise ValueError('Invalid read length with peek mode')
            return self._ftdi.read_pins()
        # in asynchronous bitbang mode, the FTDI-to-host FIFO is filled in
        # continuously once this mode is activated. This means there is no
        # way to trigger the exact moment where the buffer is filled in, nor
        # to define the write pointer in the buffer. Reading out this buffer
        # at any time is likely to contain a mix of old and new values.
        # Anyway, flushing the FTDI-to-host buffer seems to be a proper
        # to get in sync with the buffer.
        if noflush:
            return self._ftdi.read_data(readlen)
        loop = 10000
        while loop:
            loop -= 1
            # do not attempt to do anything till the FTDI HW buffer has been
            # emptied, i.e. previous write calls have been handled.
            status = self._ftdi.poll_modem_status()
            if status & Ftdi.MODEM_TEMT:
                # TX buffer is now empty, any "write" GPIO rquest has completed
                # so start reading GPIO samples from this very moment.
                break
        else:
            # sanity check to avoid endless loop on errors
            raise FtdiError('FTDI TX buffer error')
        # now flush the FTDI-to-host buffer as it keeps being filled with data
        self._ftdi.purge_tx_buffer()
        # finally perform the actual read out
        return self._ftdi.read_data(readlen)

    def write(self, out: Union[bytes, bytearray, int]) -> None:
        """Set the GPIO output pin electrical level, or output a sequence of
           bytes @ constant frequency to GPIO output pins.

           :param out: a bitfield of GPIO pins, or a sequence of them
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        if isinstance(out, (bytes, bytearray)):
            pass
        else:
            if isinstance(out, int):
                out = bytes([out])
            else:
                if not is_iterable(out):
                    raise TypeError('Invalid output value')
            for val in out:
                if val > self._mask:
                    raise ValueError('Invalid output value')
            out = bytes(out)
        self._ftdi.write_data(out)

    def set_frequency(self, frequency: Union[int, float]) -> None:
        """Set the frequency at which sequence of GPIO samples are read
           and written.

           note: FTDI may update its clock register before it has emptied its
           internal buffer. If the current frequency is "low", some
           yet-to-output bytes may end up being clocked at the new frequency.

           Unfortunately, it seems there is no way to wait for the internal
           buffer to be emptied out. They can be flushed (i.e. discarded), but
           not synchronized :-(

           PyFtdi client should add "some" short delay to ensure a previous,
           long write request has been fully output @ low freq before changing
           the frequency.

           Beware that only some exact frequencies can be generated. Contrary
           to the UART mode, an approximate frequency is always accepted for
           GPIO/bitbang mode. To get the actual frequency, and optionally abort
           if it is out-of-spec, use :py:meth:`frequency` property.

           :param frequency: the new frequency, in GPIO samples per second
        """
        self._frequency = float(self._ftdi.set_baudrate(int(frequency), False))

    def _configure(self, url: str, direction: int,
                   frequency: Union[int, float, None] = None, **kwargs) -> int:
        if 'initial' in kwargs:
            initial = kwargs['initial']
            del kwargs['initial']
        else:
            initial = None
        if 'debug' in kwargs:
            # debug is not implemented
            del kwargs['debug']
        baudrate = int(frequency) if frequency is not None else None
        baudrate = self._ftdi.open_bitbang_from_url(url,
                                                    direction=direction,
                                                    sync=False,
                                                    baudrate=baudrate,
                                                    **kwargs)
        self._width = 8
        self._mask = (1 << self._width) - 1
        self._direction = direction & self._mask
        if initial is not None:
            initial &= self._mask
            self.write(initial)
        return float(baudrate)

    def _update_direction(self) -> None:
        self._ftdi.set_bitmode(self._direction, Ftdi.BitMode.BITBANG)

    # old API names
    open_from_url = GpioBaseController.configure
    read_port = read
    write_port = write


# old API compatibility
GpioController = GpioAsyncController


class GpioSyncController(GpioBaseController):
    """GPIO controller for an FTDI port, in bit-bang synchronous mode.

       GPIO accessible pins are limited to the 8 lower pins of each GPIO port.

       Synchronous bitbang input and output are synchronized. Eveery time GPIO
       output is updated, the GPIO input is sampled and buffered.

       Update and sampling are clocked at the selected frequency. The GPIO
       samples are transfer in both direction with the :py:meth:`exchange`
       method, which therefore always returns as many input samples as output
       bytes.

       Note that FTDI internal clock divider cannot generate any arbitrary
       frequency, so the closest frequency to the request one that can be
       generated is selected. The actual :py:attr:`frequency` may be tested to
       check if it matches the board requirements.
    """

    def exchange(self, out: Union[bytes, bytearray]) -> bytes:
        """Set the GPIO output pin electrical level, or output a sequence of
           bytes @ constant frequency to GPIO output pins.

           :param out: the byte buffer to output as GPIO
           :return: a byte buffer of the same length as out buffer.
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        if isinstance(out, (bytes, bytearray)):
            pass
        else:
            if isinstance(out, int):
                out = bytes([out])
            elif not is_iterable(out):
                raise TypeError('Invalid output value')
            for val in out:
                if val > self._mask:
                    raise GpioException("Invalid value")
        self._ftdi.write_data(out)
        data = self._ftdi.read_data_bytes(len(out), 4)
        return data

    def set_frequency(self, frequency: Union[int, float]) -> None:
        """Set the frequency at which sequence of GPIO samples are read
           and written.

           :param frequency: the new frequency, in GPIO samples per second
        """
        self._frequency = float(self._ftdi.set_baudrate(int(frequency), False))

    def _configure(self, url: str, direction: int,
                   frequency: Union[int, float, None] = None, **kwargs):
        if 'initial' in kwargs:
            initial = kwargs['initial']
            del kwargs['initial']
        else:
            initial = None
        if 'debug' in kwargs:
            # debug is not implemented
            del kwargs['debug']
        baudrate = int(frequency) if frequency is not None else None
        baudrate = self._ftdi.open_bitbang_from_url(url,
                                                    direction=direction,
                                                    sync=True,
                                                    baudrate=baudrate,
                                                    **kwargs)
        self._width = 8
        self._mask = (1 << self._width) - 1
        self._direction = direction & self._mask
        if initial is not None:
            initial &= self._mask
            self.exchange(initial)
        return float(baudrate)

    def _update_direction(self) -> None:
        self._ftdi.set_bitmode(self._direction, Ftdi.BitMode.SYNCBB)


class GpioMpsseController(GpioBaseController):
    """GPIO controller for an FTDI port, in MPSSE mode.

       All GPIO pins are reachable, but MPSSE mode is slower than other modes.

       Beware that LSBs (b0..b7) and MSBs (b8..b15) are accessed with two
       subsequence commands, so a slight delay may occur when sampling or
       changing both groups at once. In other word, it is not possible to
       atomically read to / write from LSBs and MSBs. This might be worth
       checking the board design if atomic access to several lines is required.
    """

    MPSSE_PAYLOAD_MAX_LENGTH = 0xFF00  # 16 bits max (- spare for control)

    def read(self, readlen: int = 1, peek: Optional[bool] = None) -> \
             Union[int, bytes, Tuple[int]]:
        """Read the GPIO input pin electrical level.

           :param readlen: how many GPIO samples to retrieve. Each sample if
                           :py:meth:`width` bit wide.
           :param peek: whether to peak current value from port, or to use
                        MPSSE stream and HW FIFO. When peek mode is selected,
                        readlen should be 1. It is not available with wide
                        ports if some of the MSB pins are configured as input
           :return: a :py:meth:`width` bit wide integer if direct mode is used,
                    a bytes buffer if :py:meth:`width` is a byte,
                    a list of integer otherwise (MPSSE mode only).
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        if peek:
            if readlen != 1:
                raise ValueError('Invalid read length with direct mode')
            if self._width > 8:
                if (0xFFFF & ~self._direction) >> 8:
                    raise ValueError('Peek mode not available with selected '
                                     'input config')
        if peek:
            return self._ftdi.read_pins()
        return self._read_mpsse(readlen)

    def write(self, out: Union[bytes, bytearray, Iterable[int], int]) -> None:
        """Set the GPIO output pin electrical level, or output a sequence of
           bytes @ constant frequency to GPIO output pins.

           :param out: a bitfield of GPIO pins, or a sequence of them
        """
        if not self.is_connected:
            raise GpioException('Not connected')
        if isinstance(out, (bytes, bytearray)):
            pass
        else:
            if isinstance(out, int):
                out = [out]
            elif not is_iterable(out):
                raise TypeError('Invalid output value')
            for val in out:
                if val > self._mask:
                    raise GpioException("Invalid value")
        self._write_mpsse(out)

    def set_frequency(self, frequency: Union[int, float]) -> None:
        if not self.is_connected:
            raise GpioException('Not connected')
        self._frequency = self._ftdi.set_frequency(float(frequency))

    def _update_direction(self) -> None:
        # nothing to do in MPSSE mode, as direction is udpated with each
        # GPIO command
        pass

    def _configure(self, url: str, direction: int,
                   frequency: Union[int, float, None] = None, **kwargs):
        frequency = self._ftdi.open_mpsse_from_url(url,
                                                   direction=direction,
                                                   frequency=frequency,
                                                   **kwargs)
        self._width = self._ftdi.port_width
        self._mask = (1 << self._width) - 1
        self._direction = direction & self._mask
        return frequency

    def _read_mpsse(self, count: int) -> Tuple[int]:
        if self._width > 8:
            cmd = bytearray([Ftdi.GET_BITS_LOW, Ftdi.GET_BITS_HIGH] * count)
            fmt = '<%dH' % count
        else:
            cmd = bytearray([Ftdi.GET_BITS_LOW] * count)
            fmt = None
        cmd.append(Ftdi.SEND_IMMEDIATE)
        if len(cmd) > self.MPSSE_PAYLOAD_MAX_LENGTH:
            raise ValueError('Too many samples')
        self._ftdi.write_data(cmd)
        size = scalc(fmt) if fmt else count
        data = self._ftdi.read_data_bytes(size, 4)
        if len(data) != size:
            raise FtdiError('Cannot read GPIO, recv %d out of %d bytes' %
                            (len(data), size))
        if fmt:
            return sunpack(fmt, data)
        return data

    def _write_mpsse(self,
                     out: Union[bytes, bytearray, Iterable[int], int]) -> None:
        cmd = []
        low_dir = self._direction & 0xFF
        if self._width > 8:
            high_dir = (self._direction >> 8) & 0xFF
            for data in out:
                low_data = data & 0xFF
                high_data = (data >> 8) & 0xFF
                cmd.extend([Ftdi.SET_BITS_LOW, low_data, low_dir,
                            Ftdi.SET_BITS_HIGH, high_data, high_dir])
        else:
            for data in out:
                cmd.extend([Ftdi.SET_BITS_LOW, data, low_dir])
        self._ftdi.write_data(bytes(cmd))
